/**************************************************************\
* Copyright (C) 1997 - 2001 Vasile CALMATUI (vasile@wook.com)  *
*                                                              *
* This software is public domain. Permission is granted to     *
* use, copy and modify this software to anyone.                *
****************************************************************
* Please don't contact me for information, bugs, comments, job *
* offers. If you like this software, please send a postcard to *
* the following address:                                       *
* Vasile CALMATUI                                              *
* 35 str. Burebista                                            *
* 3600 Ungheni                                                 *
* MOLDOVA                                                      *
****************************************************************
* The author is not liable for damages resulting from  the use *
* of this software or its derivates.                           *
\**************************************************************/

import java.io.*;
import java.net.*;
import java.util.*;

public class VasMailAttach {
  private String mailServer="mail";

  //the line separator. Is it "\r\n" or just "\n"
  String ls=new String(System.getProperty("line.separator"));

  public String getAppletInfo() {
    return (new String("VasMailAttach v1.0 April 11th, 1999, (C) 1997, 1998, 1999, 2000, 2001 Vasile Calmatui"));
  }

  //print usage if wrong parameters were passed
  public static void printUsage() {
    die("usage: VasMailAttach [send emailFrom emailTo subject file1 file2 ..] | [read file1 file2 ..]");
  }//printUsage

  //exits the program with the specified message
  public static void die(String str) {
    System.out.println(str);
    System.exit(-1);
  }//printUsage

  //main function called in command line mode
  public static void main(String[] args) {
    VasMailAttach myClass=new VasMailAttach();
    if(args.length<1) printUsage();
    if(args[0].equals("read") && args.length>1) {
      myClass.readMail(args);
      System.exit(0);
    }
    if(args[0].equals("send") && args.length>4) {
      myClass.sendMail(args);
      System.exit(0);
    }
    printUsage();
  }//main

  private void readMail(String[] args){
    for(int i=1; i<args.length; i++) {
      try {
        RandomAccessFile file=new RandomAccessFile(args[i], "r");
        String headers;
        String body;
        //slurping the source file in the memory
        {
          byte[] fileC=new byte[(int)file.length()];
          file.readFully(fileC);
          file.close();
          String content=new String(fileC, 0);
          headers=getHeaders(content);
          if(headers==null) die("cannot find the end of the header!");
          body=getBody(content);
          if(body==null) die("cannot find the body of the mail!");
        }//end slurp
        String contentType=getHeader(headers, "Content-type:");
        if(contentType==null) die("Cannot find 'Content-type:' header!"+ls+"Maybe the message is not encoded?");
        //if the message is not multipart
        if(!contentType.toLowerCase().startsWith("multipart")) die("The type of this message is '"+contentType+"', not multipart."+ls+"Maybe the message is not encoded?");
        String boundary=getSubtype(contentType, "boundary=");
        if(boundary==null) die("Cannot find 'boundary' field in Content-type header: '"+contentType+"'!");

        //this is the beginning of the first part of the message
        int beginMultiPart=body.indexOf("--"+boundary);
        //the beginning is just after the first boundary
        if(beginMultiPart<0) die("This message has no mail parts!");
        //this is the end of the last part of the message
        int endLastMultiPart=body.indexOf("--"+boundary+"--");
        if(endLastMultiPart<0) die("Cannot find the end of encoded part of this message!");

        //this will be the new mail
        StringBuffer treatedMail=new StringBuffer(headers+ls+body.substring(0, beginMultiPart));

        int endPart=beginMultiPart;
        do {
          //divide the body of the message in parts
          endPart=body.indexOf("--"+boundary, beginMultiPart+1);
          if(endPart<0) break;
          treatedMail.append(decodePart(body.substring(beginMultiPart+2+boundary.length()+ls.length(), endPart)));
          beginMultiPart=endPart;
        }while(endPart<endLastMultiPart);
        treatedMail.append(body.substring(endLastMultiPart+2+boundary.length()+ls.length()));
        String newName=findNewFileName(args[i]);
        RandomAccessFile newMailFile=new RandomAccessFile(newName, "rw");
        newMailFile.writeBytes(treatedMail.toString());
        file.close();
        System.out.println("Your mailfile was saved in: "+newName);
      }//end try
      catch(Exception e) {
        e.printStackTrace(System.out);
        die("File problem!");
      }//end catch
    }//end for
  }//readMail

  //decodes this part of the message
  private String decodePart(String mailPart){
    String partHeader=getHeaders(mailPart);
    if(partHeader==null) System.out.println("no headers: "+mailPart);
    if(partHeader==null) return mailPart;
    String partBody=getBody(mailPart);
    if(partBody==null) return mailPart;
    String partContentTransfer=getHeader(partHeader, "Content-Transfer-Encoding:");
    if(partContentTransfer==null) return mailPart;
    //OK, will decode then it and write the output to afile
    if(partContentTransfer.toLowerCase().startsWith("base64")) {
      String contentType=getHeader(partHeader, "Content-Type:");
      String fileName=null;
      //looking if it's possible to find the "original" name of the file
      if(contentType!=null) {
        fileName=getSubtype(contentType, "name=");
      }
      //otherwise putting a default filename
      if(fileName==null) fileName=new String("attach");
      //finds a new, unique fileName
      fileName=findNewFileName(fileName);

      //writing the new file
      try {
        RandomAccessFile file=new RandomAccessFile(fileName, "rw");
        file.write(decode64(partBody));
        System.out.println("Attached file was converted to '"+fileName+"'");
        return ls+"Attached file was converted to '"+fileName+"'"+ls;
      }
      catch(Exception e) {
        e.printStackTrace(System.out);
        die(fileName+": write problem!");
      }//end catch
    }//if
    return mailPart;
  }//decodePart

  //finds a new, unique fileName
  private String findNewFileName(String name){
    int nbTest=0;
    String newName=new String(name);
    while(true) {
      if(nbTest>0) newName=new String(name+nbTest);
      File newOne=new File(newName);
      if(!newOne.exists()) return newName;
      nbTest++;
    }//while
  }//findNewFileName

  //retrieves the subtype string from a header
  private String getSubtype(String type, String subtype){
    int subtypeBegin=type.toLowerCase().indexOf(subtype);
    if(subtypeBegin<0) return null;
    StringBuffer subt=new StringBuffer(type.substring(subtypeBegin+subtype.length()).trim());
    if(subt.charAt(0)=='"') {subt.setCharAt(0, ' ');}
    if(subt.charAt(subt.length()-1)=='"') {subt.setCharAt(subt.length()-1, ' ');}
    return subt.toString().trim();
  }//getBoundary

  //retrieves the headers of a mail
  private String getHeaders(String mail){
    int headerEnd=mail.indexOf(ls+ls);
    if(headerEnd<0) return null;
    return mail.substring(0, headerEnd);
  }//getHeaders

  //retrieves the body of a mail
  private String getBody(String mail){
    int headerEnd=mail.indexOf(ls+ls);
    if(headerEnd<0) return null;
    return mail.substring(headerEnd+2*ls.length());
  }//getBody

  //retrives the given header from the mail headers
  private String getHeader(String headers, String headName){
    int headerBegin=headers.toLowerCase().indexOf(headName.toLowerCase());
    if(headerBegin<0) return null;
    int lastEnd=headerBegin+headName.length();
    StringBuffer out=new StringBuffer();
    do {
      //getting the end of the header
      int pos=headers.indexOf(ls, lastEnd);
      //if this is the last "header"
      if(pos<0) {
        out.append(headers.substring(lastEnd+1));
        break;
      }
      out.append(headers.substring(lastEnd+1, pos));
      lastEnd=pos;
      /*
        The process of moving  from  this  folded   multiple-line
        representation  of a header field to its single line represen-
        tation is called "unfolding".  Unfolding  is  accomplished  by
        regarding   CRLF   immediately  followed  by  a  LWSP-char  as
        equivalent to the LWSP-char.
      */
    }while((lastEnd+1<headers.length()) && (headers.charAt(lastEnd+1)==' ' || headers.charAt(lastEnd+1)=='\t'));
    return out.toString().trim();
  }//getHeader

  private void sendMail(String[] args){
    String boundary=new String("_www.wook.com_"+(new Date()).getTime()+"_");
    StringBuffer data=new StringBuffer("Subject: "+args[3]+"\r\nX-Mailer: VasMailAttach for Java 1.02\r\nMIME-Version: 1.0\r\nContent-type: multipart/mixed; boundary=\""+boundary+"\"\r\n\r\nThis is a multi-part message in MIME format.\r\n");

    for(int i=4; i<args.length; i++) {
      data.append("\r\n--"+boundary+"\r\nContent-Disposition: inline; filename=\""+args[i]+"\"\r\nContent-Transfer-Encoding: base64\r\nContent-Type: application/octet-stream; name=\""+args[i]+"\"\r\n\r\n");
      //name=file.getName();
      try {
        RandomAccessFile file=new RandomAccessFile(args[i], "r");
          //slurping the source file in the memory
          {
            byte[] fileC=new byte[(int)file.length()];
            file.readFully(fileC);
            file.close();
            //data.append(myClass.encode64(fileC));
            data.append(encode64(fileC));
          }//end slurp
      }//end try
      catch(Exception e) {
        System.out.println("File problem");
        e.printStackTrace(System.out);
        System.exit(-1);
      }//end catch
    }//end for
    //ending the mail
    data.append("\r\n\r\n--"+boundary+"--\r\n");
    goConnection(args[1], args[2], null, data.toString());
  }//fin main

  //the Base64 alphabet as documented in RFC 1521
  static private char[] enc_alphabet_base64=new char[64];
  //lookup table for converting base64 characters to value in range 0..63
  static private byte[] dec_alphabet_base64=new byte[256];

  static {
      for(int i=0; i<26; i++) {
          enc_alphabet_base64[i]=(char)('A'+i);
          enc_alphabet_base64[26+i]=(char)('a'+i);
      }
    for(int i=0; i<10; i++) enc_alphabet_base64[52+i]=(char)('0'+i);
    enc_alphabet_base64[62]='+';
    enc_alphabet_base64[63]='/';
    for(int i=0; i<256; i++) dec_alphabet_base64[i]=-1;
    for(int i='A'; i<='Z'; i++) dec_alphabet_base64[i]=(byte)(i-'A');
    for(int i='a'; i<='z'; i++) dec_alphabet_base64[i]=(byte)(26+i-'a');
    for(int i='0'; i<='9'; i++) dec_alphabet_base64[i]=(byte)(52+i-'0');
    dec_alphabet_base64['+']=62;
    dec_alphabet_base64['/']=63;
  }//static

  //this function makes the base64 encoding and prepares for sending
  static private String encode64(byte toEncode[]) {
    if(toEncode==null) return null;
    StringBuffer encoded=new StringBuffer();
    int i=0;
    byte a, b, c;
    for(int count=toEncode.length/3; count>0; count--) {
      //taking a "triplet"
      a=toEncode[i++];
      b=toEncode[i++];
      c=toEncode[i++];
      encoded.append(enc_alphabet_base64[(a >>> 2) & 0x3F]);
      encoded.append(enc_alphabet_base64[((a << 4) & 0x30) + ((b >>> 4) & 0xf)]);
      encoded.append(enc_alphabet_base64[((b << 2) & 0x3c) + ((c >>> 6) & 0x3)]);
      encoded.append(enc_alphabet_base64[c & 0x3F]);
      //add an newline after every 76 chars=57/3*4
      if (i!=0 && i%57==0) encoded.append("\r\n");
    }//for
    //treating "the remainings"
    if(toEncode.length%3==1) {
      a=toEncode[i++];
      encoded.append(enc_alphabet_base64[(a >>> 2) & 0x3F]);
      encoded.append(enc_alphabet_base64[((a << 4) & 0x30)]);
      encoded.append("==");
    }
    if(toEncode.length%3==2) {
      a=toEncode[i++];
      b=toEncode[i++];
      encoded.append(enc_alphabet_base64[(a >>> 2) & 0x3F]);
      encoded.append(enc_alphabet_base64[((a << 4) & 0x30) + ((b >>> 4) & 0xf)]);
      encoded.append(enc_alphabet_base64[((b << 2) & 0x3c)]);
      encoded.append('=');
    }
    encoded.append("\r\n");
    return encoded.toString();
  }//encode64

//this function makes the base64 decoding
static private byte[] decode64(String toDecode) {
    int strLen=toDecode.length();
    char[] goodChars=toDecode.toCharArray();
    int nbGoodChars=0;
    for(int i=0; i<strLen; i++) {
        //Ignores whitespace and newlines
        if(goodChars[i]!=' ' && goodChars[i]!='\r' && goodChars[i]!='\n') goodChars[nbGoodChars++]=goodChars[i];
    }//for

    int len=((nbGoodChars + 3) / 4) * 3;
    if(nbGoodChars>0 && goodChars[nbGoodChars-1]=='=') --len;
    if(nbGoodChars>1 && goodChars[nbGoodChars-2]=='=') --len;
    byte[] out=new byte[len];
    int shift=0;   // # of excess bits stored in accum
    int accum=0;   // excess bits
    int index=0;
    for(int ix=0; ix<nbGoodChars; ix++) {
        //ignore high byte of char
        int value=dec_alphabet_base64[goodChars[ix] & 0xFF];
        //skip over non-code
        if(value>=0) {
            //bits shift up by 6 each time thru
            //loop, with new bits being put in
            // at the bottom.
            accum<<=6;
            shift+=6;
            accum|=value;
            //whenever there are 8 or more shifted in
            //write them out (from the top, leaving any
            //excess at the bottom for next iteration
            if(shift>=8) {
                shift-=8;
                out[index++]=(byte)((accum >> shift) & 0xff);
            }//if
        }//if
    }//for
    if(index!=out.length) throw new Error("miscalculated data length!");
    return out;
}//decode64

  //reading the reply from the server
  boolean readReply(DataInputStream in) {
    String line="";
    try {
      line=in.readLine();
    }
    catch(IOException e) {System.out.println("Reading reply problem"); return true;}
    if(line.startsWith("4") || line.startsWith("5")) {
      System.out.println("Err: "+line);
      return true;
    }
    else return false;
  }//end readReply

  public void goConnection(String from, String to, String subj, String data) {
    try {
      System.out.println("Establishing connection..");
      Socket sock=new Socket(mailServer, 25);
      DataOutputStream out=new DataOutputStream(sock.getOutputStream());
      DataInputStream in=new DataInputStream(sock.getInputStream());
      System.out.println("Connection OK, sending data");

      if(readReply(in)) {sock.close(); return;}
      out.writeBytes("HELO vasile25.md\r\n");
      if(readReply(in)) {sock.close(); return;}
      out.writeBytes("MAIL FROM:<"+from+">\r\n");
      if(readReply(in)) {sock.close(); return;}
      out.writeBytes("RCPT TO:<"+to+">\r\n");
      if(readReply(in)) {sock.close(); return;}
      out.writeBytes("DATA\r\n");
      if(readReply(in)) {sock.close(); return;}
      if(subj!=null) out.writeBytes("Subject: "+subj+"\n\n");
      out.writeBytes(data+"\r\n.\r\nQUIT\r\n");
      if(readReply(in)) {sock.close(); return;}
      sock.close();
      System.out.println("E-mail with attachementsent.");
    }
    catch(IOException e) {System.out.println("Connection problem or connection closed");}
  };//end goConnection
}//end class VasSendMail
